#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.genai.types import (
    Content,
    CreateCachedContentConfig,
    EmbedContentConfig,
    GoogleSearch,
    Part,
    Tool,
)

from airflow.providers.google.cloud.hooks.gen_ai import (
    GenAIGeminiAPIHook,
    GenAIGenerativeModelHook,
)

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
)

TEST_GCP_CONN_ID: str = "test-gcp-conn-id"
GCP_PROJECT = "test-project"
GCP_LOCATION = "us-central1"

TEST_PROMPT = "In 10 words or less, what is apache airflow?"
TEST_CONTENTS = [TEST_PROMPT]
TEST_LANGUAGE_PRETRAINED_MODEL = "textembedding-gecko"
TEST_TEMPERATURE = 0.0
TEST_MAX_OUTPUT_TOKENS = 256
TEST_TOP_P = 0.8
TEST_TOP_K = 40

TEST_TEXT_EMBEDDING_MODEL = ""
TEST_TEXT_EMBEDDING_CONFIG = EmbedContentConfig(output_dimensionality=10)

TEST_MULTIMODAL_PRETRAINED_MODEL = "gemini-pro"

TEST_GENERATION_CONFIG = {
    "max_output_tokens": TEST_MAX_OUTPUT_TOKENS,
    "top_p": TEST_TOP_P,
    "temperature": TEST_TEMPERATURE,
}
TEST_TOOLS = [Tool(google_search=GoogleSearch())]

TEST_MULTIMODAL_VISION_MODEL = "gemini-pro-vision"

SOURCE_MODEL = "gemini-1.0-pro-002"
TRAIN_DATASET = "gs://cloud-samples-data/ai-platform/generative_ai/sft_train_data.jsonl"

TEST_CACHED_MODEL = "gemini-1.5-pro-002"
TEST_CACHED_SYSTEM_INSTRUCTION = """
You are an expert researcher. You always stick to the facts in the sources provided, and never make up new facts.
Now look at these research papers, and answer the following questions.
"""

CACHED_CONTENT_CONFIG = CreateCachedContentConfig(
    contents=[
        Content(
            role="user",
            parts=[
                Part.from_uri(
                    file_uri="gs://cloud-samples-data/generative-ai/pdf/2312.11805v3.pdf",
                    mime_type="application/pdf",
                ),
                Part.from_uri(
                    file_uri="gs://cloud-samples-data/generative-ai/pdf/2403.05530.pdf",
                    mime_type="application/pdf",
                ),
            ],
        )
    ],
    system_instruction=TEST_CACHED_SYSTEM_INSTRUCTION,
    display_name="test-cache",
    ttl="3600s",
)

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
GENERATIVE_MODEL_STRING = "airflow.providers.google.cloud.hooks.gen_ai.{}"

TEST_API_KEY = "test-api-key"
TEST_JOB_NAME = "batches/test-job-id"
TEST_MODEL = "models/gemini-2.5-flash"
TEST_BATCH_JOB_SOURCE_INLINE = [
    {"contents": [{"parts": [{"text": "Tell me a one-sentence joke."}], "role": "user"}]},
    {"contents": [{"parts": [{"text": "Why is the sky blue?"}], "role": "user"}]},
]
TEST_EMBEDDINGS_JOB_SOURCE_INLINE = {
    "contents": [{"parts": [{"text": "Why is the sky blue?"}], "role": "user"}]
}
TEST_SOURCE_FILE = "test-bucket/source.jsonl"
TEST_LOCAL_FILE_PATH = "/tmp/data/test_file.json"
TEST_FILE_NAME = "files/test-file-id"

# Mock constants for configuration objects
TEST_LIST_BATCH_JOBS_CONFIG = {"page_size": 10}
TEST_CREATE_BATCH_JOB_CONFIG = {"display_name": "test-job"}
TEST_UPLOAD_FILE_CONFIG = {"display_name": "custom_name", "mime_type": "text/plain"}


def assert_warning(msg: str, warnings):
    assert any(msg in str(w) for w in warnings)


class TestGenAIGenerativeModelHookWithDefaultProjectId:
    def dummy_get_credentials(self):
        pass

    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = GenAIGenerativeModelHook(gcp_conn_id=TEST_GCP_CONN_ID)
            self.hook.get_credentials = self.dummy_get_credentials

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_text_embedding_model_get_embeddings(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.embed_content(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=TEST_CONTENTS,
            model=TEST_TEXT_EMBEDDING_MODEL,
            config=TEST_TEXT_EMBEDDING_CONFIG,
        )
        client_mock.models.embed_content.assert_called_once_with(
            model=TEST_TEXT_EMBEDDING_MODEL,
            contents=TEST_CONTENTS,
            config=TEST_TEXT_EMBEDDING_CONFIG,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_generative_model_generate_content(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.generate_content(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=TEST_CONTENTS,
            generation_config=TEST_GENERATION_CONFIG,
            model=TEST_MULTIMODAL_PRETRAINED_MODEL,
        )
        client_mock.models.generate_content.assert_called_once_with(
            model=TEST_MULTIMODAL_PRETRAINED_MODEL,
            contents=TEST_CONTENTS,
            config=TEST_GENERATION_CONFIG,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_supervised_fine_tuning_train(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.supervised_fine_tuning_train(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            source_model=SOURCE_MODEL,
            training_dataset=TRAIN_DATASET,
        )
        client_mock.tunings.tune.assert_called_once_with(
            base_model=SOURCE_MODEL,
            training_dataset=TRAIN_DATASET,
            config=None,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_count_tokens(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.count_tokens(
            project_id=GCP_PROJECT,
            contents=TEST_CONTENTS,
            location=GCP_LOCATION,
            model=TEST_MULTIMODAL_PRETRAINED_MODEL,
        )
        client_mock.models.count_tokens.assert_called_once_with(
            model=TEST_MULTIMODAL_PRETRAINED_MODEL,
            contents=TEST_CONTENTS,
            config=None,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_create_cached_content(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.create_cached_content(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=TEST_CACHED_MODEL,
            cached_content_config=CACHED_CONTENT_CONFIG,
        )
        client_mock.caches.create.assert_called_once_with(
            model=TEST_CACHED_MODEL,
            config=CACHED_CONTENT_CONFIG,
        )


class TestGenAIGeminiAPIHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = GenAIGeminiAPIHook(gemini_api_key=TEST_API_KEY)

    @mock.patch("google.genai.Client")
    def test_get_genai_client(self, mock_client):
        """Test client initialization with correct parameters."""
        self.hook.get_genai_client()

        mock_client.assert_called_once_with(
            api_key=TEST_API_KEY,
            vertexai=False,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_get_batch_job(self, mock_get_client):
        client_mock = mock_get_client.return_value
        client_mock.batches = mock.Mock()

        self.hook.get_batch_job(job_name=TEST_JOB_NAME)

        client_mock.batches.get.assert_called_once_with(name=TEST_JOB_NAME)

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_list_batch_jobs(self, mock_get_client):
        client_mock = mock_get_client.return_value
        client_mock.batches = mock.Mock()

        self.hook.list_batch_jobs(list_batch_jobs_config=TEST_LIST_BATCH_JOBS_CONFIG)

        client_mock.batches.list.assert_called_once_with(config=TEST_LIST_BATCH_JOBS_CONFIG)

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_create_batch_job(self, mock_get_client):
        client_mock = mock_get_client.return_value
        client_mock.batches = mock.Mock()

        self.hook.create_batch_job(
            model=TEST_MODEL,
            source=TEST_BATCH_JOB_SOURCE_INLINE,
            create_batch_job_config=TEST_CREATE_BATCH_JOB_CONFIG,
        )

        client_mock.batches.create.assert_called_once_with(
            model=TEST_MODEL, src=TEST_BATCH_JOB_SOURCE_INLINE, config=TEST_CREATE_BATCH_JOB_CONFIG
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_delete_batch_job(self, mock_get_client):
        client_mock = mock_get_client.return_value
        client_mock.batches = mock.Mock()

        self.hook.delete_batch_job(job_name=TEST_JOB_NAME)

        client_mock.batches.delete.assert_called_once_with(name=TEST_JOB_NAME)

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_cancel_batch_job(self, mock_get_client):
        client_mock = mock_get_client.return_value
        client_mock.batches = mock.Mock()

        self.hook.cancel_batch_job(job_name=TEST_JOB_NAME)

        client_mock.batches.cancel.assert_called_once_with(name=TEST_JOB_NAME)

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_create_embeddings_with_inline_source(self, mock_get_client):
        """Test create_embeddings when source is a dict (inline)."""
        client_mock = mock_get_client.return_value
        client_mock.batches = mock.Mock()

        self.hook.create_embeddings(
            model=TEST_MODEL,
            source=TEST_EMBEDDINGS_JOB_SOURCE_INLINE,
            create_embeddings_config=TEST_CREATE_BATCH_JOB_CONFIG,
        )

        client_mock.batches.create_embeddings.assert_called_once_with(
            model=TEST_MODEL,
            src={"inlined_requests": TEST_EMBEDDINGS_JOB_SOURCE_INLINE},
            config=TEST_CREATE_BATCH_JOB_CONFIG,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_create_embeddings_with_file_source(self, mock_get_client):
        """Test create_embeddings when source is a string (file name)."""
        client_mock = mock_get_client.return_value
        client_mock.batches = mock.Mock()

        # Test with str (File name)
        source_file = "/bucket/file.jsonl"
        self.hook.create_embeddings(
            model=TEST_MODEL, source=source_file, create_embeddings_config=TEST_CREATE_BATCH_JOB_CONFIG
        )

        client_mock.batches.create_embeddings.assert_called_once_with(
            model=TEST_MODEL, src={"file_name": source_file}, config=TEST_CREATE_BATCH_JOB_CONFIG
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_upload_file_with_provided_config(self, mock_get_client):
        """Test upload_file when explicit config is provided."""
        client_mock = mock_get_client.return_value
        client_mock.files = mock.Mock()

        self.hook.upload_file(path_to_file=TEST_LOCAL_FILE_PATH, upload_file_config=TEST_UPLOAD_FILE_CONFIG)

        client_mock.files.upload.assert_called_once_with(
            file=TEST_LOCAL_FILE_PATH, config=TEST_UPLOAD_FILE_CONFIG
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_upload_file_default_config_generation(self, mock_get_client):
        """Test that upload_file generates correct config from filename if config is None."""
        client_mock = mock_get_client.return_value
        client_mock.files = mock.Mock()

        # Path: /tmp/data/test_file.json -> name: test_file, type: json
        self.hook.upload_file(path_to_file=TEST_LOCAL_FILE_PATH, upload_file_config=None)

        expected_config = {"display_name": "test_file", "mime_type": "json"}

        client_mock.files.upload.assert_called_once_with(file=TEST_LOCAL_FILE_PATH, config=expected_config)

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_get_file(self, mock_get_client):
        client_mock = mock_get_client.return_value
        client_mock.files = mock.Mock()

        self.hook.get_file(file_name=TEST_FILE_NAME)

        client_mock.files.get.assert_called_once_with(name=TEST_FILE_NAME)

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_download_file(self, mock_get_client):
        client_mock = mock_get_client.return_value
        client_mock.files = mock.Mock()

        self.hook.download_file(file_name=TEST_FILE_NAME)

        client_mock.files.download.assert_called_once_with(file=TEST_FILE_NAME)

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_list_files(self, mock_get_client):
        client_mock = mock_get_client.return_value
        client_mock.files = mock.Mock()

        self.hook.list_files()

        client_mock.files.list.assert_called_once()

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGeminiAPIHook.get_genai_client"))
    def test_delete_file(self, mock_get_client):
        client_mock = mock_get_client.return_value
        client_mock.files = mock.Mock()

        self.hook.delete_file(file_name=TEST_FILE_NAME)

        client_mock.files.delete.assert_called_once_with(name=TEST_FILE_NAME)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Unit Tests for the GSheets Hook
"""

from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.suite.hooks.sheets import GSheetsHook

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

GCP_CONN_ID = "test"
SPREADSHEET_ID = "1234567890"
RANGE_ = "test!A:E"
RANGES = ["test!A:Q", "test!R:Z"]
VALUES = [[1, 2, 3]]
VALUES_BATCH = [[[1, 2, 3]], [[4, 5, 6]]]
MAJOR_DIMENSION = "ROWS"
VALUE_RENDER_OPTION = "FORMATTED_VALUE"
DATE_TIME_RENDER_OPTION = "SERIAL_NUMBER"
INCLUDE_VALUES_IN_RESPONSE = True
VALUE_INPUT_OPTION = "RAW"
INSERT_DATA_OPTION = "OVERWRITE"
NUM_RETRIES = 5
API_RESPONSE = {"test": "response"}


class TestGSheetsHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = GSheetsHook(gcp_conn_id=GCP_CONN_ID)

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook._authorize")
    @mock.patch("airflow.providers.google.suite.hooks.sheets.build")
    def test_gsheets_client_creation(self, mock_build, mock_authorize):
        result = self.hook.get_conn()
        mock_build.assert_called_once()
        args, kwargs = mock_build.call_args
        assert kwargs["http"] == mock_authorize.return_value
        assert kwargs["cache_discovery"] is False
        assert mock_build.return_value == result

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook._authorize")
    @mock.patch("airflow.providers.google.suite.hooks.sheets.build")
    def test_gsheets_hook_custom_endpoint(self, mock_build, mock_authorize):
        self.hook.api_endpoint = "https://private.googleapis.com"
        self.hook.get_conn()
        mock_build.assert_called_once()
        _, kwargs = mock_build.call_args
        client_options = kwargs.get("client_options")
        if client_options is None:
            api_endpoint = None
        else:
            api_endpoint = getattr(client_options, "api_endpoint", None)

        assert api_endpoint == "https://private.googleapis.com"

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_get_values(self, get_conn):
        get_method = get_conn.return_value.spreadsheets.return_value.values.return_value.get
        execute_method = get_method.return_value.execute
        execute_method.return_value = {"values": VALUES}
        result = self.hook.get_values(
            spreadsheet_id=SPREADSHEET_ID,
            range_=RANGE_,
            major_dimension=MAJOR_DIMENSION,
            value_render_option=VALUE_RENDER_OPTION,
            date_time_render_option=DATE_TIME_RENDER_OPTION,
        )
        assert result is VALUES
        execute_method.assert_called_once_with(num_retries=NUM_RETRIES)
        get_method.assert_called_once_with(
            spreadsheetId=SPREADSHEET_ID,
            range=RANGE_,
            majorDimension=MAJOR_DIMENSION,
            valueRenderOption=VALUE_RENDER_OPTION,
            dateTimeRenderOption=DATE_TIME_RENDER_OPTION,
        )

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_get_values_empty(self, get_conn):
        get_method = get_conn.return_value.spreadsheets.return_value.values.return_value.get
        execute_method = get_method.return_value.execute
        execute_method.return_value = {}
        result = self.hook.get_values(
            spreadsheet_id=SPREADSHEET_ID,
            range_=RANGE_,
            major_dimension=MAJOR_DIMENSION,
            value_render_option=VALUE_RENDER_OPTION,
            date_time_render_option=DATE_TIME_RENDER_OPTION,
        )
        assert result == []
        execute_method.assert_called_once_with(num_retries=NUM_RETRIES)
        get_method.assert_called_once_with(
            spreadsheetId=SPREADSHEET_ID,
            range=RANGE_,
            majorDimension=MAJOR_DIMENSION,
            valueRenderOption=VALUE_RENDER_OPTION,
            dateTimeRenderOption=DATE_TIME_RENDER_OPTION,
        )

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_batch_get_values(self, get_conn):
        batch_get_method = get_conn.return_value.spreadsheets.return_value.values.return_value.batchGet
        execute_method = batch_get_method.return_value.execute
        execute_method.return_value = API_RESPONSE
        result = self.hook.batch_get_values(
            spreadsheet_id=SPREADSHEET_ID,
            ranges=RANGES,
            major_dimension=MAJOR_DIMENSION,
            value_render_option=VALUE_RENDER_OPTION,
            date_time_render_option=DATE_TIME_RENDER_OPTION,
        )
        assert result is API_RESPONSE
        execute_method.assert_called_once_with(num_retries=NUM_RETRIES)
        batch_get_method.assert_called_once_with(
            spreadsheetId=SPREADSHEET_ID,
            ranges=RANGES,
            majorDimension=MAJOR_DIMENSION,
            valueRenderOption=VALUE_RENDER_OPTION,
            dateTimeRenderOption=DATE_TIME_RENDER_OPTION,
        )

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_update_values(self, get_conn):
        update_method = get_conn.return_value.spreadsheets.return_value.values.return_value.update
        execute_method = update_method.return_value.execute
        execute_method.return_value = API_RESPONSE
        result = self.hook.update_values(
            spreadsheet_id=SPREADSHEET_ID,
            range_=RANGE_,
            values=VALUES,
            major_dimension=MAJOR_DIMENSION,
            value_input_option=VALUE_INPUT_OPTION,
            include_values_in_response=INCLUDE_VALUES_IN_RESPONSE,
            value_render_option=VALUE_RENDER_OPTION,
            date_time_render_option=DATE_TIME_RENDER_OPTION,
        )
        body = {"range": RANGE_, "majorDimension": MAJOR_DIMENSION, "values": VALUES}
        assert result is API_RESPONSE
        execute_method.assert_called_once_with(num_retries=NUM_RETRIES)
        update_method.assert_called_once_with(
            spreadsheetId=SPREADSHEET_ID,
            range=RANGE_,
            valueInputOption=VALUE_INPUT_OPTION,
            includeValuesInResponse=INCLUDE_VALUES_IN_RESPONSE,
            responseValueRenderOption=VALUE_RENDER_OPTION,
            responseDateTimeRenderOption=DATE_TIME_RENDER_OPTION,
            body=body,
        )

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_batch_update_values(self, get_conn):
        batch_update_method = get_conn.return_value.spreadsheets.return_value.values.return_value.batchUpdate
        execute_method = batch_update_method.return_value.execute
        execute_method.return_value = API_RESPONSE
        result = self.hook.batch_update_values(
            spreadsheet_id=SPREADSHEET_ID,
            ranges=RANGES,
            values=VALUES_BATCH,
            major_dimension=MAJOR_DIMENSION,
            value_input_option=VALUE_INPUT_OPTION,
            include_values_in_response=INCLUDE_VALUES_IN_RESPONSE,
            value_render_option=VALUE_RENDER_OPTION,
            date_time_render_option=DATE_TIME_RENDER_OPTION,
        )
        data = []
        for idx, range_ in enumerate(RANGES):
            value_range = {"range": range_, "majorDimension": MAJOR_DIMENSION, "values": VALUES_BATCH[idx]}
            data.append(value_range)
        body = {
            "valueInputOption": VALUE_INPUT_OPTION,
            "data": data,
            "includeValuesInResponse": INCLUDE_VALUES_IN_RESPONSE,
            "responseValueRenderOption": VALUE_RENDER_OPTION,
            "responseDateTimeRenderOption": DATE_TIME_RENDER_OPTION,
        }
        assert result is API_RESPONSE
        execute_method.assert_called_once_with(num_retries=NUM_RETRIES)
        batch_update_method.assert_called_once_with(spreadsheetId=SPREADSHEET_ID, body=body)

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_batch_update_values_with_bad_data(self, get_conn):
        batch_update_method = get_conn.return_value.spreadsheets.return_value.values.return_value.batchUpdate
        execute_method = batch_update_method.return_value.execute
        execute_method.return_value = API_RESPONSE
        with pytest.raises(AirflowException) as ctx:
            self.hook.batch_update_values(
                spreadsheet_id=SPREADSHEET_ID,
                ranges=["test!A1:B2", "test!C1:C2"],
                values=[[1, 2, 3]],  # bad data
                major_dimension=MAJOR_DIMENSION,
                value_input_option=VALUE_INPUT_OPTION,
                include_values_in_response=INCLUDE_VALUES_IN_RESPONSE,
                value_render_option=VALUE_RENDER_OPTION,
                date_time_render_option=DATE_TIME_RENDER_OPTION,
            )
        batch_update_method.assert_not_called()
        execute_method.assert_not_called()
        err = ctx.value
        assert "must be of equal length." in str(err)

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_append_values(self, get_conn):
        append_method = get_conn.return_value.spreadsheets.return_value.values.return_value.append
        execute_method = append_method.return_value.execute
        execute_method.return_value = API_RESPONSE
        result = self.hook.append_values(
            spreadsheet_id=SPREADSHEET_ID,
            range_=RANGE_,
            values=VALUES,
            major_dimension=MAJOR_DIMENSION,
            value_input_option=VALUE_INPUT_OPTION,
            insert_data_option=INSERT_DATA_OPTION,
            include_values_in_response=INCLUDE_VALUES_IN_RESPONSE,
            value_render_option=VALUE_RENDER_OPTION,
            date_time_render_option=DATE_TIME_RENDER_OPTION,
        )
        body = {"range": RANGE_, "majorDimension": MAJOR_DIMENSION, "values": VALUES}
        assert result is API_RESPONSE
        execute_method.assert_called_once_with(num_retries=NUM_RETRIES)
        append_method.assert_called_once_with(
            spreadsheetId=SPREADSHEET_ID,
            range=RANGE_,
            valueInputOption=VALUE_INPUT_OPTION,
            insertDataOption=INSERT_DATA_OPTION,
            includeValuesInResponse=INCLUDE_VALUES_IN_RESPONSE,
            responseValueRenderOption=VALUE_RENDER_OPTION,
            responseDateTimeRenderOption=DATE_TIME_RENDER_OPTION,
            body=body,
        )

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_clear_values(self, get_conn):
        clear_method = get_conn.return_value.spreadsheets.return_value.values.return_value.clear
        execute_method = clear_method.return_value.execute
        execute_method.return_value = API_RESPONSE
        result = self.hook.clear(spreadsheet_id=SPREADSHEET_ID, range_=RANGE_)

        assert result is API_RESPONSE
        execute_method.assert_called_once_with(num_retries=NUM_RETRIES)
        clear_method.assert_called_once_with(spreadsheetId=SPREADSHEET_ID, range=RANGE_)

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_batch_clear_values(self, get_conn):
        batch_clear_method = get_conn.return_value.spreadsheets.return_value.values.return_value.batchClear
        execute_method = batch_clear_method.return_value.execute
        execute_method.return_value = API_RESPONSE
        result = self.hook.batch_clear(spreadsheet_id=SPREADSHEET_ID, ranges=RANGES)
        body = {"ranges": RANGES}
        assert result is API_RESPONSE
        execute_method.assert_called_once_with(num_retries=NUM_RETRIES)
        batch_clear_method.assert_called_once_with(spreadsheetId=SPREADSHEET_ID, body=body)

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_get_spreadsheet(self, mock_get_conn):
        get_mock = mock_get_conn.return_value.spreadsheets.return_value.get
        get_mock.return_value.execute.return_value = API_RESPONSE

        result = self.hook.get_spreadsheet(spreadsheet_id=SPREADSHEET_ID)

        get_mock.assert_called_once_with(spreadsheetId=SPREADSHEET_ID)
        assert result == API_RESPONSE

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_spreadsheet")
    def test_get_sheet_titles(self, mock_get_spreadsheet):
        sheet1 = {"properties": {"title": "title1"}}
        sheet2 = {"properties": {"title": "title2"}}
        mock_get_spreadsheet.return_value = {"sheets": [sheet1, sheet2]}

        result = self.hook.get_sheet_titles(spreadsheet_id=SPREADSHEET_ID)
        mock_get_spreadsheet.assert_called_once_with(spreadsheet_id=SPREADSHEET_ID)
        assert result == ["title1", "title2"]

        result = self.hook.get_sheet_titles(spreadsheet_id=SPREADSHEET_ID, sheet_filter=["title1"])
        assert result == ["title1"]

    @mock.patch("airflow.providers.google.suite.hooks.sheets.GSheetsHook.get_conn")
    def test_create_spreadsheet(self, mock_get_conn):
        spreadsheet = mock.MagicMock()

        create_mock = mock_get_conn.return_value.spreadsheets.return_value.create
        create_mock.return_value.execute.return_value = API_RESPONSE

        result = self.hook.create_spreadsheet(spreadsheet=spreadsheet)

        create_mock.assert_called_once_with(body=spreadsheet)
        assert result == API_RESPONSE



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from copy import deepcopy

import pytest

from airflow.providers.google.cloud.utils.field_sanitizer import GcpBodyFieldSanitizer


class TestGcpBodyFieldSanitizer:
    def test_sanitize_should_sanitize_empty_body_and_fields(self):
        body = {}
        fields_to_sanitize = []

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body == {}

    def test_sanitize_should_not_fail_with_none_body(self):
        body = None
        fields_to_sanitize = []

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body is None

    def test_sanitize_should_fail_with_none_fields(self):
        body = {}
        fields_to_sanitize = None

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)

        with pytest.raises(TypeError):
            sanitizer.sanitize(body)

    def test_sanitize_should_not_fail_if_field_is_absent_in_body(self):
        body = {}
        fields_to_sanitize = ["kind"]

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body == {}

    def test_sanitize_should_not_remove_fields_for_incorrect_specification(self):
        actual_body = [
            {"kind": "compute#instanceTemplate", "name": "instance"},
            {"kind": "compute#instanceTemplate1", "name": "instance1"},
            {"kind": "compute#instanceTemplate2", "name": "instance2"},
        ]
        body = deepcopy(actual_body)
        fields_to_sanitize = ["kind"]

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert actual_body == body

    def test_sanitize_should_remove_all_fields_from_root_level(self):
        body = {"kind": "compute#instanceTemplate", "name": "instance"}
        fields_to_sanitize = ["kind"]

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body == {"name": "instance"}

    def test_sanitize_should_remove_for_multiple_fields_from_root_level(self):
        body = {"kind": "compute#instanceTemplate", "name": "instance"}
        fields_to_sanitize = ["kind", "name"]

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body == {}

    def test_sanitize_should_remove_all_fields_in_a_list_value(self):
        body = {
            "fields": [
                {"kind": "compute#instanceTemplate", "name": "instance"},
                {"kind": "compute#instanceTemplate1", "name": "instance1"},
                {"kind": "compute#instanceTemplate2", "name": "instance2"},
            ]
        }
        fields_to_sanitize = ["fields.kind"]

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body == {
            "fields": [
                {"name": "instance"},
                {"name": "instance1"},
                {"name": "instance2"},
            ]
        }

    def test_sanitize_should_remove_all_fields_in_any_nested_body(self):
        fields_to_sanitize = [
            "kind",
            "properties.disks.kind",
            "properties.metadata.kind",
        ]

        body = {
            "kind": "compute#instanceTemplate",
            "name": "instance",
            "properties": {
                "disks": [
                    {
                        "name": "a",
                        "kind": "compute#attachedDisk",
                        "type": "PERSISTENT",
                        "mode": "READ_WRITE",
                    },
                    {
                        "name": "b",
                        "kind": "compute#attachedDisk",
                        "type": "PERSISTENT",
                        "mode": "READ_WRITE",
                    },
                ],
                "metadata": {"kind": "compute#metadata", "fingerprint": "GDPUYxlwHe4="},
            },
        }
        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body == {
            "name": "instance",
            "properties": {
                "disks": [
                    {"name": "a", "type": "PERSISTENT", "mode": "READ_WRITE"},
                    {"name": "b", "type": "PERSISTENT", "mode": "READ_WRITE"},
                ],
                "metadata": {"fingerprint": "GDPUYxlwHe4="},
            },
        }

    def test_sanitize_should_not_fail_if_specification_has_none_value(self):
        fields_to_sanitize = [
            "kind",
            "properties.disks.kind",
            "properties.metadata.kind",
        ]

        body = {"kind": "compute#instanceTemplate", "name": "instance", "properties": {"disks": None}}

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body == {"name": "instance", "properties": {"disks": None}}

    def test_sanitize_should_not_fail_if_no_specification_matches(self):
        fields_to_sanitize = [
            "properties.disks.kind1",
            "properties.metadata.kind2",
        ]

        body = {"name": "instance", "properties": {"disks": None}}

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body == {"name": "instance", "properties": {"disks": None}}

    def test_sanitize_should_not_fail_if_type_in_body_do_not_match_with_specification(self):
        fields_to_sanitize = [
            "properties.disks.kind",
            "properties.metadata.kind2",
        ]

        body = {"name": "instance", "properties": {"disks": 1}}

        sanitizer = GcpBodyFieldSanitizer(fields_to_sanitize)
        sanitizer.sanitize(body)

        assert body == {"name": "instance", "properties": {"disks": 1}}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that shows how to use DataFusion.
"""

from __future__ import annotations

import os
from datetime import datetime

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.dag import DAG
from airflow.providers.google.cloud.hooks.datafusion import DataFusionHook
from airflow.providers.google.cloud.operators.datafusion import (
    CloudDataFusionCreateInstanceOperator,
    CloudDataFusionCreatePipelineOperator,
    CloudDataFusionDeleteInstanceOperator,
    CloudDataFusionDeletePipelineOperator,
    CloudDataFusionGetInstanceOperator,
    CloudDataFusionListPipelinesOperator,
    CloudDataFusionRestartInstanceOperator,
    CloudDataFusionStartPipelineOperator,
    CloudDataFusionStopPipelineOperator,
    CloudDataFusionUpdateInstanceOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.sensors.datafusion import CloudDataFusionPipelineStateSensor

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

# [START howto_data_fusion_env_variables]
SERVICE_ACCOUNT = os.environ.get("GCP_DATAFUSION_SERVICE_ACCOUNT")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
LOCATION = "europe-north1"
DAG_ID = "datafusion"
INSTANCE_NAME = f"df-{ENV_ID}".replace("_", "-")
INSTANCE = {
    "type": "BASIC",
    "displayName": INSTANCE_NAME,
    "dataprocServiceAccount": SERVICE_ACCOUNT,
}

BUCKET_NAME_1 = f"bucket1-{DAG_ID}-{ENV_ID}".replace("_", "-")
BUCKET_NAME_2 = f"bucket2-{DAG_ID}-{ENV_ID}".replace("_", "-")
BUCKET_NAME_1_URI = f"gs://{BUCKET_NAME_1}"
BUCKET_NAME_2_URI = f"gs://{BUCKET_NAME_2}"

PIPELINE_NAME = f"pipe-{ENV_ID}".replace("_", "-")
PIPELINE = {
    "artifact": {
        "name": "cdap-data-pipeline",
        "version": "{{ task_instance.xcom_pull(task_ids='get_artifacts_versions')['cdap-data-pipeline'] }}",
        "scope": "SYSTEM",
    },
    "description": "Data Pipeline Application",
    "name": PIPELINE_NAME,
    "config": {
        "resources": {"memoryMB": 2048, "virtualCores": 1},
        "driverResources": {"memoryMB": 2048, "virtualCores": 1},
        "connections": [{"from": "GCS", "to": "GCS2"}],
        "comments": [],
        "postActions": [],
        "properties": {},
        "processTimingEnabled": "true",
        "stageLoggingEnabled": "false",
        "stages": [
            {
                "name": "GCS",
                "plugin": {
                    "name": "GCSFile",
                    "type": "batchsource",
                    "label": "GCS",
                    "artifact": {
                        "name": "google-cloud",
                        "version": "{{ task_instance.xcom_pull(task_ids='get_artifacts_versions')\
                            ['google-cloud'] }}",
                        "scope": "SYSTEM",
                    },
                    "properties": {
                        "project": "auto-detect",
                        "format": "text",
                        "skipHeader": "false",
                        "serviceFilePath": "auto-detect",
                        "filenameOnly": "false",
                        "recursive": "false",
                        "encrypted": "false",
                        "schema": '{"type":"record","name":"textfile","fields":[{"name"\
                            :"offset","type":"long"},{"name":"body","type":"string"}]}',
                        "path": BUCKET_NAME_1_URI,
                        "referenceName": "foo_bucket",
                        "useConnection": "false",
                        "serviceAccountType": "filePath",
                        "sampleSize": "1000",
                        "fileEncoding": "UTF-8",
                    },
                },
                "outputSchema": '{"type":"record","name":"textfile","fields"\
                    :[{"name":"offset","type":"long"},{"name":"body","type":"string"}]}',
                "id": "GCS",
            },
            {
                "name": "GCS2",
                "plugin": {
                    "name": "GCS",
                    "type": "batchsink",
                    "label": "GCS2",
                    "artifact": {
                        "name": "google-cloud",
                        "version": "{{ task_instance.xcom_pull(task_ids='get_artifacts_versions')\
                            ['google-cloud'] }}",
                        "scope": "SYSTEM",
                    },
                    "properties": {
                        "project": "auto-detect",
                        "suffix": "yyyy-MM-dd-HH-mm",
                        "format": "json",
                        "serviceFilePath": "auto-detect",
                        "location": "us",
                        "schema": '{"type":"record","name":"textfile","fields":[{"name"\
                            :"offset","type":"long"},{"name":"body","type":"string"}]}',
                        "referenceName": "bar",
                        "path": BUCKET_NAME_2_URI,
                        "serviceAccountType": "filePath",
                        "contentType": "application/octet-stream",
                    },
                },
                "outputSchema": '{"type":"record","name":"textfile","fields"\
                    :[{"name":"offset","type":"long"},{"name":"body","type":"string"}]}',
                "inputSchema": [
                    {
                        "name": "GCS",
                        "schema": '{"type":"record","name":"textfile","fields":[{"name"\
                            :"offset","type":"long"},{"name":"body","type":"string"}]}',
                    }
                ],
                "id": "GCS2",
            },
        ],
        "schedule": "0 * * * *",
        "engine": "spark",
        "numOfRecordsPreview": 100,
        "description": "Data Pipeline Application",
        "maxConcurrentRuns": 1,
    },
}
# [END howto_data_fusion_env_variables]

CloudDataFusionCreatePipelineOperator.template_fields = (
    *CloudDataFusionCreatePipelineOperator.template_fields,
    "pipeline",
)

with DAG(
    DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule="@once",
    catchup=False,
    tags=["example", "datafusion"],
) as dag:
    create_bucket1 = GCSCreateBucketOperator(
        task_id="create_bucket1",
        bucket_name=BUCKET_NAME_1,
        project_id=PROJECT_ID,
    )

    create_bucket2 = GCSCreateBucketOperator(
        task_id="create_bucket2",
        bucket_name=BUCKET_NAME_2,
        project_id=PROJECT_ID,
    )

    # [START howto_cloud_data_fusion_create_instance_operator]
    create_instance = CloudDataFusionCreateInstanceOperator(
        location=LOCATION,
        instance_name=INSTANCE_NAME,
        instance=INSTANCE,
        task_id="create_instance",
    )
    # [END howto_cloud_data_fusion_create_instance_operator]

    # [START howto_cloud_data_fusion_get_instance_operator]
    get_instance = CloudDataFusionGetInstanceOperator(
        location=LOCATION, instance_name=INSTANCE_NAME, task_id="get_instance"
    )
    # [END howto_cloud_data_fusion_get_instance_operator]

    # [START howto_cloud_data_fusion_restart_instance_operator]
    restart_instance = CloudDataFusionRestartInstanceOperator(
        location=LOCATION, instance_name=INSTANCE_NAME, task_id="restart_instance"
    )
    # [END howto_cloud_data_fusion_restart_instance_operator]

    # [START howto_cloud_data_fusion_update_instance_operator]
    update_instance = CloudDataFusionUpdateInstanceOperator(
        location=LOCATION,
        instance_name=INSTANCE_NAME,
        instance=INSTANCE,
        update_mask="",
        task_id="update_instance",
    )
    # [END howto_cloud_data_fusion_update_instance_operator]

    @task(task_id="get_artifacts_versions")
    def get_artifacts_versions(ti=None):
        hook = DataFusionHook()
        instance_url = ti.xcom_pull(task_ids="get_instance", key="return_value")["apiEndpoint"]
        artifacts = hook.get_instance_artifacts(instance_url=instance_url, namespace="default")
        return {item["name"]: item["version"] for item in artifacts}

    # [START howto_cloud_data_fusion_create_pipeline]
    create_pipeline = CloudDataFusionCreatePipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        pipeline=PIPELINE,
        instance_name=INSTANCE_NAME,
        task_id="create_pipeline",
    )
    # [END howto_cloud_data_fusion_create_pipeline]

    # [START howto_cloud_data_fusion_list_pipelines]
    list_pipelines = CloudDataFusionListPipelinesOperator(
        location=LOCATION, instance_name=INSTANCE_NAME, task_id="list_pipelines"
    )
    # [END howto_cloud_data_fusion_list_pipelines]

    # [START howto_cloud_data_fusion_start_pipeline]
    start_pipeline = CloudDataFusionStartPipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        instance_name=INSTANCE_NAME,
        pipeline_timeout=1000,
        task_id="start_pipeline",
    )
    # [END howto_cloud_data_fusion_start_pipeline]

    # [START howto_cloud_data_fusion_start_pipeline_def]
    start_pipeline_def = CloudDataFusionStartPipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        instance_name=INSTANCE_NAME,
        task_id="start_pipeline_def",
        deferrable=True,
    )
    # [END howto_cloud_data_fusion_start_pipeline_def]

    # [START howto_cloud_data_fusion_start_pipeline_async]
    start_pipeline_async = CloudDataFusionStartPipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        instance_name=INSTANCE_NAME,
        asynchronous=True,
        task_id="start_pipeline_async",
    )
    # [END howto_cloud_data_fusion_start_pipeline_async]

    # [START howto_cloud_data_fusion_start_pipeline_sensor]
    start_pipeline_sensor = CloudDataFusionPipelineStateSensor(
        task_id="pipeline_state_sensor",
        pipeline_name=PIPELINE_NAME,
        pipeline_id=start_pipeline_async.output,
        expected_statuses=["COMPLETED"],
        failure_statuses=["FAILED"],
        instance_name=INSTANCE_NAME,
        location=LOCATION,
    )
    # [END howto_cloud_data_fusion_start_pipeline_sensor]

    # [START howto_cloud_data_fusion_stop_pipeline]
    stop_pipeline = CloudDataFusionStopPipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        instance_name=INSTANCE_NAME,
        task_id="stop_pipeline",
    )
    # [END howto_cloud_data_fusion_stop_pipeline]

    # [START howto_cloud_data_fusion_delete_pipeline]
    delete_pipeline = CloudDataFusionDeletePipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        instance_name=INSTANCE_NAME,
        task_id="delete_pipeline",
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_cloud_data_fusion_delete_pipeline]

    # [START howto_cloud_data_fusion_delete_instance_operator]
    delete_instance = CloudDataFusionDeleteInstanceOperator(
        location=LOCATION,
        instance_name=INSTANCE_NAME,
        task_id="delete_instance",
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_cloud_data_fusion_delete_instance_operator]

    delete_bucket1 = GCSDeleteBucketOperator(
        task_id="delete_bucket1", bucket_name=BUCKET_NAME_1, trigger_rule=TriggerRule.ALL_DONE
    )
    delete_bucket2 = GCSDeleteBucketOperator(
        task_id="delete_bucket2", bucket_name=BUCKET_NAME_1, trigger_rule=TriggerRule.ALL_DONE
    )

    (
        # TEST SETUP
        [create_bucket1, create_bucket2]
        # TEST BODY
        >> create_instance
        >> get_instance
        >> get_artifacts_versions()
        >> restart_instance
        >> update_instance
        >> create_pipeline
        >> list_pipelines
        >> start_pipeline_def
        >> start_pipeline_async
        >> start_pipeline_sensor
        >> start_pipeline
        >> stop_pipeline
        >> delete_pipeline
        >> delete_instance
        # TEST TEARDOWN
        >> [delete_bucket1, delete_bucket2]
    )

    from tests_common.test_utils.watcher import watcher

# This test needs watcher in order to properly mark success/failure
# when "tearDown" task with trigger rule is part of the DAG
list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.cloud.transfers.sheets_to_gcs import GoogleSheetsToGCSOperator

RANGE = "test!A:E"
FILTER = ["sheet_filter"]
SHEET_TITLE = "title"
RANGES = ["test1", "test2"]
SPREADSHEET_ID = "1234567890"
VALUES = [[1, 2, 3]]
BUCKET = "destination_bucket"
PATH = "path/to/reports"
GCP_CONN_ID = "test"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]


class TestGoogleSheetsToGCSOperator:
    @mock.patch("airflow.providers.google.cloud.transfers.sheets_to_gcs.csv.writer")
    @mock.patch("airflow.providers.google.cloud.transfers.sheets_to_gcs.NamedTemporaryFile")
    def test_upload_data(self, mock_tempfile, mock_writer):
        filename = "file://97g23r"
        file_handle = mock.MagicMock()
        mock_tempfile.return_value.__enter__.return_value = file_handle
        mock_tempfile.return_value.__enter__.return_value.name = filename

        mock_sheet_hook = mock.MagicMock()
        mock_sheet_hook.get_spreadsheet.return_value = {"properties": {"title": SHEET_TITLE}}
        expected_dest_file = f"{PATH}/{SHEET_TITLE}_{RANGE}.csv"

        mock_gcs_hook = mock.MagicMock()

        op = GoogleSheetsToGCSOperator(
            task_id="test_task",
            spreadsheet_id=SPREADSHEET_ID,
            destination_bucket=BUCKET,
            sheet_filter=FILTER,
            destination_path=PATH,
        )

        result = op._upload_data(
            gcs_hook=mock_gcs_hook,
            hook=mock_sheet_hook,
            sheet_range=RANGE,
            sheet_values=VALUES,
        )

        # Test writing to file
        mock_sheet_hook.get_spreadsheet.assert_called_once_with(SPREADSHEET_ID)
        mock_writer.assert_called_once_with(file_handle)
        mock_writer.return_value.writerows.assert_called_once_with(VALUES)
        file_handle.flush.assert_called_once_with()

        # Test upload
        mock_gcs_hook.upload.assert_called_once_with(
            bucket_name=BUCKET, object_name=expected_dest_file, filename=filename
        )

        # Assert path to file is returned
        assert result == expected_dest_file

    @mock.patch("airflow.providers.google.cloud.transfers.sheets_to_gcs.GCSHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sheets_to_gcs.GSheetsHook")
    @mock.patch(
        "airflow.providers.google.cloud.transfers.sheets_to_gcs.GoogleSheetsToGCSOperator._upload_data"
    )
    def test_execute(self, mock_upload_data, mock_sheet_hook, mock_gcs_hook):
        mock_ti = mock.MagicMock()
        mock_context = {"ti": mock_ti}
        data = ["data1", "data2"]
        mock_sheet_hook.return_value.get_sheet_titles.return_value = RANGES
        mock_sheet_hook.return_value.get_values.side_effect = data
        mock_upload_data.side_effect = [PATH, PATH]

        op = GoogleSheetsToGCSOperator(
            task_id="test_task",
            spreadsheet_id=SPREADSHEET_ID,
            destination_bucket=BUCKET,
            sheet_filter=FILTER,
            destination_path=PATH,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(mock_context)

        mock_sheet_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_gcs_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_sheet_hook.return_value.get_sheet_titles.assert_called_once_with(
            spreadsheet_id=SPREADSHEET_ID, sheet_filter=FILTER
        )

        calls = [mock.call(spreadsheet_id=SPREADSHEET_ID, range_=r) for r in RANGES]
        mock_sheet_hook.return_value.get_values.assert_has_calls(calls)

        calls = [
            mock.call(mock_gcs_hook.return_value, mock_sheet_hook.return_value, r, v)
            for r, v in zip(RANGES, data)
        ]
        mock_upload_data.assert_has_calls(calls)
        actual_call_count = mock_upload_data.call_count
        assert len(RANGES) == actual_call_count

        mock_ti.xcom_push.assert_called_once_with(key="destination_objects", value=[PATH, PATH])



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.google.cloud.links.dataplex import (
    DataplexCatalogAspectTypeLink,
    DataplexCatalogAspectTypesLink,
    DataplexCatalogEntryGroupLink,
    DataplexCatalogEntryGroupsLink,
    DataplexCatalogEntryLink,
    DataplexCatalogEntryTypeLink,
    DataplexCatalogEntryTypesLink,
    DataplexLakeLink,
    DataplexTaskLink,
    DataplexTasksLink,
)
from airflow.providers.google.cloud.operators.dataplex import (
    DataplexCatalogCreateAspectTypeOperator,
    DataplexCatalogCreateEntryGroupOperator,
    DataplexCatalogCreateEntryTypeOperator,
    DataplexCatalogGetAspectTypeOperator,
    DataplexCatalogGetEntryGroupOperator,
    DataplexCatalogGetEntryOperator,
    DataplexCatalogGetEntryTypeOperator,
    DataplexCreateLakeOperator,
    DataplexCreateTaskOperator,
    DataplexListTasksOperator,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk.execution_time.comms import XComResult

TEST_LOCATION = "test-location"
TEST_PROJECT_ID = "test-project-id"
TEST_ENTRY_ID = "test-entry-id"
TEST_ENTRY_GROUP_ID = "test-entry-group-id"
TEST_ENTRY_GROUP_ID_BODY = {"description": "some description"}
TEST_ENTRY_GROUPS_ID = "test-entry-groups-id"
TEST_ENTRY_TYPE_ID = "test-entry-type-id"
TEST_ENTRY_TYPE_ID_BODY = {"description": "some description"}
TEST_ENTRY_TYPES_ID = "test-entry-types-id"
TEST_ASPECT_TYPE_ID = "test-aspect-type-id"
TEST_ASPECT_TYPE_ID_BODY = {"description": "some description"}
TEST_ASPECT_TYPES_ID = "test-aspect-types-id"
TEST_TASK_ID = "test-task-id"
TEST_TASKS_ID = "test-tasks-id"
TEST_LAKE_ID = "test-lake-id"
TEST_LAKE_BODY = {"name": "some_name"}

DATAPLEX_BASE_LINK = "https://console.cloud.google.com/dataplex/"
EXPECTED_DATAPLEX_CATALOG_ENTRY_GROUP_LINK = (
    DATAPLEX_BASE_LINK
    + f"projects/{TEST_PROJECT_ID}/locations/{TEST_LOCATION}/entryGroups/{TEST_ENTRY_GROUP_ID}?project={TEST_PROJECT_ID}"
)
EXPECTED_DATAPLEX_CATALOG_ENTRY_GROUPS_LINK = (
    DATAPLEX_BASE_LINK + f"catalog/entry-groups?project={TEST_PROJECT_ID}"
)
EXPECTED_DATAPLEX_CATALOG_ENTRY_TYPE_LINK = (
    DATAPLEX_BASE_LINK
    + f"projects/{TEST_PROJECT_ID}/locations/{TEST_LOCATION}/entryTypes/{TEST_ENTRY_TYPE_ID}?project={TEST_PROJECT_ID}"
)
EXPECTED_DATAPLEX_CATALOG_ENTRY_TYPES_LINK = (
    DATAPLEX_BASE_LINK + f"catalog/entry-types?project={TEST_PROJECT_ID}"
)
DATAPLEX_LAKE_LINK = (
    DATAPLEX_BASE_LINK + f"lakes/{TEST_LAKE_ID};location={TEST_LOCATION}?project={TEST_PROJECT_ID}"
)
EXPECTED_DATAPLEX_TASK_LINK = (
    DATAPLEX_BASE_LINK
    + f"process/tasks/{TEST_LAKE_ID}.{TEST_TASK_ID};location={TEST_LOCATION}/jobs?project={TEST_PROJECT_ID}"
)
EXPECTED_DATAPLEX_TASKS_LINK = (
    DATAPLEX_BASE_LINK + f"process/tasks?project={TEST_PROJECT_ID}&qLake={TEST_LAKE_ID}.{TEST_LOCATION}"
)
EXPECTED_DATAPLEX_CATALOG_ASPECT_TYPE_LINK = (
    DATAPLEX_BASE_LINK
    + f"projects/{TEST_PROJECT_ID}/locations/{TEST_LOCATION}/aspectTypes/{TEST_ASPECT_TYPE_ID}?project={TEST_PROJECT_ID}"
)
EXPECTED_DATAPLEX_CATALOG_ASPECT_TYPES_LINK = (
    DATAPLEX_BASE_LINK + f"catalog/aspect-types?project={TEST_PROJECT_ID}"
)
EXPECTED_DATAPLEX_CATALOG_ENTRY_LINK = (
    DATAPLEX_BASE_LINK
    + f"dp-entries/projects/{TEST_PROJECT_ID}/locations/{TEST_LOCATION}/entryGroups/{TEST_ENTRY_GROUP_ID}/entries/{TEST_ENTRY_ID}?project={TEST_PROJECT_ID}"
)


class TestDataplexTaskLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_DATAPLEX_TASK_LINK
        link = DataplexTaskLink()
        ti = create_task_instance_of_operator(
            DataplexCreateTaskOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            region=TEST_LOCATION,
            lake_id=TEST_LAKE_ID,
            project_id=TEST_PROJECT_ID,
            body=TEST_LAKE_BODY,
            dataplex_task_id=TEST_TASK_ID,
        )
        task = dag_maker.dag.get_task(ti.task_id)

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "lake_id": task.lake_id,
                    "task_id": task.dataplex_task_id,
                    "region": task.region,
                    "project_id": task.project_id,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestDataplexTasksLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_DATAPLEX_TASKS_LINK
        link = DataplexTasksLink()
        ti = create_task_instance_of_operator(
            DataplexListTasksOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            region=TEST_LOCATION,
            lake_id=TEST_LAKE_ID,
            project_id=TEST_PROJECT_ID,
        )
        task = dag_maker.dag.get_task(ti.task_id)

        link.persist(context={"ti": ti, "task": task})

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "project_id": task.project_id,
                    "lake_id": task.lake_id,
                    "region": task.region,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestDataplexLakeLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = DATAPLEX_LAKE_LINK
        link = DataplexLakeLink()
        ti = create_task_instance_of_operator(
            DataplexCreateLakeOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            region=TEST_LOCATION,
            lake_id=TEST_LAKE_ID,
            project_id=TEST_PROJECT_ID,
            body={},
        )
        task = dag_maker.dag.get_task(ti.task_id)

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "lake_id": task.lake_id,
                    "region": task.region,
                    "project_id": task.project_id,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestDataplexCatalogEntryGroupLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_DATAPLEX_CATALOG_ENTRY_GROUP_LINK
        link = DataplexCatalogEntryGroupLink()
        ti = create_task_instance_of_operator(
            DataplexCatalogGetEntryGroupOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            location=TEST_LOCATION,
            entry_group_id=TEST_ENTRY_GROUP_ID,
            project_id=TEST_PROJECT_ID,
        )
        task = dag_maker.dag.get_task(ti.task_id)

        link.persist(context={"ti": ti, "task": task})

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "entry_group_id": task.entry_group_id,
                    "location": task.location,
                    "project_id": task.project_id,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestDataplexCatalogEntryGroupsLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_DATAPLEX_CATALOG_ENTRY_GROUPS_LINK
        link = DataplexCatalogEntryGroupsLink()
        ti = create_task_instance_of_operator(
            DataplexCatalogCreateEntryGroupOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            location=TEST_LOCATION,
            entry_group_id=TEST_ENTRY_GROUP_ID,
            entry_group_configuration=TEST_ENTRY_GROUP_ID_BODY,
            project_id=TEST_PROJECT_ID,
        )
        task = dag_maker.dag.get_task(ti.task_id)

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "location": task.location,
                    "project_id": task.project_id,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestDataplexCatalogEntryTypeLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_DATAPLEX_CATALOG_ENTRY_TYPE_LINK
        link = DataplexCatalogEntryTypeLink()
        ti = create_task_instance_of_operator(
            DataplexCatalogGetEntryTypeOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            location=TEST_LOCATION,
            entry_type_id=TEST_ENTRY_TYPE_ID,
            project_id=TEST_PROJECT_ID,
        )
        task = dag_maker.dag.get_task(ti.task_id)

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "entry_type_id": task.entry_type_id,
                    "location": task.location,
                    "project_id": task.project_id,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestDataplexCatalogEntryTypesLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_DATAPLEX_CATALOG_ENTRY_TYPES_LINK
        link = DataplexCatalogEntryTypesLink()
        ti = create_task_instance_of_operator(
            DataplexCatalogCreateEntryTypeOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            location=TEST_LOCATION,
            entry_type_id=TEST_ENTRY_TYPE_ID,
            entry_type_configuration=TEST_ENTRY_TYPE_ID_BODY,
            project_id=TEST_PROJECT_ID,
        )
        task = dag_maker.dag.get_task(ti.task_id)

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "location": task.location,
                    "project_id": task.project_id,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestDataplexCatalogAspectTypeLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_DATAPLEX_CATALOG_ASPECT_TYPE_LINK
        link = DataplexCatalogAspectTypeLink()
        ti = create_task_instance_of_operator(
            DataplexCatalogGetAspectTypeOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            location=TEST_LOCATION,
            aspect_type_id=TEST_ASPECT_TYPE_ID,
            project_id=TEST_PROJECT_ID,
        )
        task = dag_maker.dag.get_task(ti.task_id)

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "aspect_type_id": task.aspect_type_id,
                    "location": task.location,
                    "project_id": task.project_id,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestDataplexCatalogAspectTypesLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_DATAPLEX_CATALOG_ASPECT_TYPES_LINK
        link = DataplexCatalogAspectTypesLink()
        ti = create_task_instance_of_operator(
            DataplexCatalogCreateAspectTypeOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            location=TEST_LOCATION,
            aspect_type_id=TEST_ASPECT_TYPE_ID,
            aspect_type_configuration=TEST_ASPECT_TYPE_ID_BODY,
            project_id=TEST_PROJECT_ID,
        )
        task = dag_maker.dag.get_task(ti.task_id)

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "location": task.location,
                    "project_id": task.project_id,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestDataplexCatalogEntryLink:
    @pytest.mark.db_test
    def test_get_link(self, dag_maker, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_DATAPLEX_CATALOG_ENTRY_LINK
        link = DataplexCatalogEntryLink()
        ti = create_task_instance_of_operator(
            DataplexCatalogGetEntryOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            location=TEST_LOCATION,
            entry_id=TEST_ENTRY_ID,
            entry_group_id=TEST_ENTRY_GROUP_ID,
            project_id=TEST_PROJECT_ID,
        )
        task = dag_maker.dag.get_task(ti.task_id)

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "entry_id": task.entry_id,
                    "entry_group_id": task.entry_group_id,
                    "location": task.location,
                    "project_id": task.project_id,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Vertex AI operators."""

from __future__ import annotations

from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any

from google.api_core.exceptions import NotFound
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.aiplatform_v1 import types

from airflow.providers.common.compat.sdk import AirflowException, conf
from airflow.providers.google.cloud.hooks.vertex_ai.pipeline_job import PipelineJobHook
from airflow.providers.google.cloud.links.vertex_ai import (
    VertexAIPipelineJobLink,
    VertexAIPipelineJobListLink,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator
from airflow.providers.google.cloud.triggers.vertex_ai import RunPipelineJobTrigger

if TYPE_CHECKING:
    from google.api_core.retry import Retry
    from google.cloud.aiplatform import PipelineJob
    from google.cloud.aiplatform.metadata import experiment_resources

    from airflow.providers.common.compat.sdk import Context


class RunPipelineJobOperator(GoogleCloudBaseOperator):
    """
    Create and run a Pipeline job.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param display_name: Required. The user-defined name of this Pipeline.
    :param template_path: Required. The path of PipelineJob or PipelineSpec JSON or YAML file. It can be
        a local path, a Google Cloud Storage URI (e.g. "gs://project.name"), an Artifact Registry URI
        (e.g. "https://us-central1-kfp.pkg.dev/proj/repo/pack/latest"), or an HTTPS URI.
    :param job_id: Optional. The unique ID of the job run. If not specified, pipeline name + timestamp
        will be used.
    :param pipeline_root: Optional. The root of the pipeline outputs. If not set, the staging bucket set
        in aiplatform.init will be used. If that's not set a pipeline-specific artifacts bucket will be
        used.
    :param parameter_values: Optional. The mapping from runtime parameter names to its values that
        control the pipeline run.
    :param input_artifacts: Optional. The mapping from the runtime parameter name for this artifact to
        its resource id. For example: "vertex_model":"456". Note: full resource name
        ("projects/123/locations/us-central1/metadataStores/default/artifacts/456") cannot be used.
    :param enable_caching: Optional. Whether to turn on caching for the run.
        If this is not set, defaults to the compile time settings, which are True for all tasks by
        default, while users may specify different caching options for individual tasks.
        If this is set, the setting applies to all tasks in the pipeline. Overrides the compile time
        settings.
    :param encryption_spec_key_name: Optional. The Cloud KMS resource identifier of the customer managed
        encryption key used to protect the job. Has the form:
        ``projects/my-project/locations/my-region/keyRings/my-kr/cryptoKeys/my-key``.
        The key needs to be in the same region as where the compute resource is created. If this is set,
        then all resources created by the PipelineJob will be encrypted with the provided encryption key.
        Overrides encryption_spec_key_name set in aiplatform.init.
    :param labels: Optional. The user defined metadata to organize PipelineJob.
    :param failure_policy: Optional. The failure policy - "slow" or "fast". Currently, the default of a
        pipeline is that the pipeline will continue to run until no more tasks can be executed, also
        known as PIPELINE_FAILURE_POLICY_FAIL_SLOW (corresponds to "slow"). However, if a pipeline is set
        to PIPELINE_FAILURE_POLICY_FAIL_FAST (corresponds to "fast"), it will stop scheduling any new
        tasks when a task has failed. Any scheduled tasks will continue to completion.
    :param service_account: Optional. Specifies the service account for workload run-as account. Users
        submitting jobs must have act-as permission on this run-as account.
    :param network: Optional. The full name of the Compute Engine network to which the job should be
        peered. For example, projects/12345/global/networks/myVPC.
        Private services access must already be configured for the network. If left unspecified, the
        network set in aiplatform.init will be used. Otherwise, the job is not peered with any network.
    :param create_request_timeout: Optional. The timeout for the create request in seconds.
    :param experiment: Optional. The Vertex AI experiment name or instance to associate to this PipelineJob.
        Metrics produced by the PipelineJob as system.Metric Artifacts will be associated as metrics
        to the current Experiment Run. Pipeline parameters will be associated as parameters to
        the current Experiment Run.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: If True, run the task in the deferrable mode.
    :param poll_interval: Time (seconds) to wait between two consecutive calls to check the job.
        The default is 300 seconds.
    """

    template_fields = [
        "region",
        "project_id",
        "input_artifacts",
        "impersonation_chain",
        "template_path",
        "pipeline_root",
        "parameter_values",
        "service_account",
    ]
    operator_extra_links = (VertexAIPipelineJobLink(),)

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        display_name: str,
        template_path: str,
        job_id: str | None = None,
        pipeline_root: str | None = None,
        parameter_values: dict[str, Any] | None = None,
        input_artifacts: dict[str, str] | None = None,
        enable_caching: bool | None = None,
        encryption_spec_key_name: str | None = None,
        labels: dict[str, str] | None = None,
        failure_policy: str | None = None,
        service_account: str | None = None,
        network: str | None = None,
        create_request_timeout: float | None = None,
        experiment: str | experiment_resources.Experiment | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 5 * 60,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.display_name = display_name
        self.template_path = template_path
        self.job_id = job_id
        self.pipeline_root = pipeline_root
        self.parameter_values = parameter_values
        self.input_artifacts = input_artifacts
        self.enable_caching = enable_caching
        self.encryption_spec_key_name = encryption_spec_key_name
        self.labels = labels
        self.failure_policy = failure_policy
        self.service_account = service_account
        self.network = network
        self.create_request_timeout = create_request_timeout
        self.experiment = experiment
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.poll_interval = poll_interval

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "region": self.region,
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        self.log.info("Running Pipeline job")
        pipeline_job_obj: PipelineJob = self.hook.submit_pipeline_job(
            project_id=self.project_id,
            region=self.region,
            display_name=self.display_name,
            template_path=self.template_path,
            job_id=self.job_id,
            pipeline_root=self.pipeline_root,
            parameter_values=self.parameter_values,
            input_artifacts=self.input_artifacts,
            enable_caching=self.enable_caching,
            encryption_spec_key_name=self.encryption_spec_key_name,
            labels=self.labels,
            failure_policy=self.failure_policy,
            service_account=self.service_account,
            network=self.network,
            create_request_timeout=self.create_request_timeout,
            experiment=self.experiment,
        )
        pipeline_job_id = pipeline_job_obj.job_id
        self.log.info("Pipeline job was created. Job id: %s", pipeline_job_id)
        context["ti"].xcom_push(key="pipeline_job_id", value=pipeline_job_id)
        VertexAIPipelineJobLink.persist(context=context, pipeline_id=pipeline_job_id)

        if self.deferrable:
            pipeline_job_obj.wait_for_resource_creation()
            self.defer(
                trigger=RunPipelineJobTrigger(
                    conn_id=self.gcp_conn_id,
                    project_id=self.project_id,
                    location=pipeline_job_obj.location,
                    job_id=pipeline_job_id,
                    poll_interval=self.poll_interval,
                    impersonation_chain=self.impersonation_chain,
                ),
                method_name="execute_complete",
            )

        pipeline_job_obj.wait()
        pipeline_job = pipeline_job_obj.to_dict()
        return pipeline_job

    def execute_complete(self, context: Context, event: dict[str, Any]) -> dict[str, Any]:
        if event["status"] == "error":
            raise AirflowException(event["message"])
        return event["job"]

    def on_kill(self) -> None:
        """Act as a callback called when the operator is killed; cancel any running job."""
        if self.hook:
            self.hook.cancel_pipeline_job()

    @cached_property
    def hook(self) -> PipelineJobHook:
        return PipelineJobHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )


class GetPipelineJobOperator(GoogleCloudBaseOperator):
    """
    Get a Pipeline job.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param pipeline_job_id: Required. The ID of the PipelineJob resource.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).

    """

    template_fields = [
        "region",
        "pipeline_job_id",
        "project_id",
        "impersonation_chain",
    ]
    operator_extra_links = (VertexAIPipelineJobLink(),)

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        pipeline_job_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.pipeline_job_id = pipeline_job_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "region": self.region,
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        hook = PipelineJobHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        try:
            self.log.info("Get Pipeline job: %s", self.pipeline_job_id)
            result = hook.get_pipeline_job(
                project_id=self.project_id,
                region=self.region,
                pipeline_job_id=self.pipeline_job_id,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
            VertexAIPipelineJobLink.persist(context=context, pipeline_id=self.pipeline_job_id)
            self.log.info("Pipeline job was gotten.")
            return types.PipelineJob.to_dict(result)
        except NotFound:
            self.log.info("The Pipeline job %s does not exist.", self.pipeline_job_id)


class ListPipelineJobOperator(GoogleCloudBaseOperator):
    """
    Lists PipelineJob in a Location.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param filter: Optional. Lists the PipelineJobs that match the filter expression. The
        following fields are supported:

        -  ``pipeline_name``: Supports ``=`` and ``!=`` comparisons.
        -  ``display_name``: Supports ``=``, ``!=`` comparisons, and
            ``:`` wildcard.
        -  ``pipeline_job_user_id``: Supports ``=``, ``!=``
            comparisons, and ``:`` wildcard. for example, can check
            if pipeline's display_name contains *step* by doing
            display_name:"*step*"
        -  ``create_time``: Supports ``=``, ``!=``, ``<``, ``>``,
            ``<=``, and ``>=`` comparisons. Values must be in RFC
            3339 format.
        -  ``update_time``: Supports ``=``, ``!=``, ``<``, ``>``,
            ``<=``, and ``>=`` comparisons. Values must be in RFC
            3339 format.
        -  ``end_time``: Supports ``=``, ``!=``, ``<``, ``>``,
            ``<=``, and ``>=`` comparisons. Values must be in RFC
            3339 format.
        -  ``labels``: Supports key-value equality and key presence.

        Filter expressions can be combined together using logical
        operators (``AND`` & ``OR``). For example:
        ``pipeline_name="test" AND create_time>"2020-05-18T13:30:00Z"``.

        The syntax to define filter expression is based on
        https://google.aip.dev/160.
    :param page_size: Optional. The standard list page size.
    :param page_token: Optional. The standard list page token. Typically obtained via
        [ListPipelineJobsResponse.next_page_token][google.cloud.aiplatform.v1.ListPipelineJobsResponse.next_page_token]
        of the previous
        [PipelineService.ListPipelineJobs][google.cloud.aiplatform.v1.PipelineService.ListPipelineJobs]
        call.
    :param order_by: Optional. A comma-separated list of fields to order by. The default
        sort order is in ascending order. Use "desc" after a field
        name for descending. You can have multiple order_by fields
        provided e.g. "create_time desc, end_time", "end_time,
        start_time, update_time" For example, using "create_time
        desc, end_time" will order results by create time in
        descending order, and if there are multiple jobs having the
        same create time, order them by the end time in ascending
        order. if order_by is not specified, it will order by
        default order is create time in descending order. Supported
        fields:

        -  ``create_time``
        -  ``update_time``
        -  ``end_time``
        -  ``start_time``
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = [
        "region",
        "project_id",
        "impersonation_chain",
    ]
    operator_extra_links = [
        VertexAIPipelineJobListLink(),
    ]

    def __init__(
        self,
        *,
        region: str,
        project_id: str,
        page_size: int | None = None,
        page_token: str | None = None,
        filter: str | None = None,
        order_by: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.page_size = page_size
        self.page_token = page_token
        self.filter = filter
        self.order_by = order_by
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "region": self.region,
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        hook = PipelineJobHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        results = hook.list_pipeline_jobs(
            region=self.region,
            project_id=self.project_id,
            page_size=self.page_size,
            page_token=self.page_token,
            filter=self.filter,
            order_by=self.order_by,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        VertexAIPipelineJobListLink.persist(context=context)
        return [types.PipelineJob.to_dict(result) for result in results]


class DeletePipelineJobOperator(GoogleCloudBaseOperator):
    """
    Delete a Pipeline job.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param pipeline_job_id: Required. The ID of the PipelineJob resource to be deleted.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = [
        "region",
        "project_id",
        "pipeline_job_id",
        "impersonation_chain",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        pipeline_job_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.pipeline_job_id = pipeline_job_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = PipelineJobHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        try:
            self.log.info("Deleting Pipeline job: %s", self.pipeline_job_id)
            operation = hook.delete_pipeline_job(
                region=self.region,
                project_id=self.project_id,
                pipeline_job_id=self.pipeline_job_id,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
            hook.wait_for_operation(timeout=self.timeout, operation=operation)
            self.log.info("Pipeline job was deleted.")
        except NotFound:
            self.log.info("The Pipeline Job ID %s does not exist.", self.pipeline_job_id)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import google.api_core.exceptions
import pytest
from google.cloud.bigtable.instance import Instance
from google.cloud.bigtable.table import ClusterState

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.sensors.bigtable import BigtableTableReplicationCompletedSensor

PROJECT_ID = "test_project_id"
INSTANCE_ID = "test-instance-id"
GCP_CONN_ID = "test-gcp-conn-id"
TABLE_ID = "test-table-id"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]


class BigtableWaitForTableReplicationTest:
    @pytest.mark.parametrize(
        ("missing_attribute", "project_id", "instance_id", "table_id"),
        [
            ("instance_id", PROJECT_ID, "", TABLE_ID),
            ("table_id", PROJECT_ID, INSTANCE_ID, ""),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.sensors.bigtable.BigtableHook")
    def test_empty_attribute(self, missing_attribute, project_id, instance_id, table_id, mock_hook):
        with pytest.raises(AirflowException) as ctx:
            BigtableTableReplicationCompletedSensor(
                project_id=project_id,
                instance_id=instance_id,
                table_id=table_id,
                task_id="id",
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )
        err = ctx.value
        assert str(err) == f"Empty parameter: {missing_attribute}"
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.sensors.bigtable.BigtableHook")
    def test_wait_no_instance(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None

        op = BigtableTableReplicationCompletedSensor(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        assert not op.poke(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

    @mock.patch("airflow.providers.google.cloud.sensors.bigtable.BigtableHook")
    def test_wait_no_table(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)
        mock_hook.return_value.get_cluster_states_for_table.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.NotFound("Table not found.")
        )

        op = BigtableTableReplicationCompletedSensor(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        assert not op.poke(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

    @mock.patch("airflow.providers.google.cloud.sensors.bigtable.BigtableHook")
    def test_wait_not_ready(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)
        mock_hook.return_value.get_cluster_states_for_table.return_value = {"cl-id": ClusterState(0)}
        op = BigtableTableReplicationCompletedSensor(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        assert not op.poke(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

    @mock.patch("airflow.providers.google.cloud.sensors.bigtable.BigtableHook")
    def test_wait_ready(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)
        mock_hook.return_value.get_cluster_states_for_table.return_value = {"cl-id": ClusterState(4)}
        op = BigtableTableReplicationCompletedSensor(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        assert op.poke(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module contains a Google Cloud Vertex AI hook.

.. spelling:word-list::

    aiplatform
"""

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from google.api_core.client_options import ClientOptions
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.aiplatform import PipelineJob
from google.cloud.aiplatform_v1 import (
    PipelineServiceAsyncClient,
    PipelineServiceClient,
    PipelineState,
    types,
)

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import GoogleBaseAsyncHook, GoogleBaseHook
from airflow.providers.google.common.hooks.operation_helpers import OperationHelper

if TYPE_CHECKING:
    from google.api_core.operation import Operation
    from google.api_core.retry import AsyncRetry, Retry
    from google.auth.credentials import Credentials
    from google.cloud.aiplatform.metadata import experiment_resources
    from google.cloud.aiplatform_v1.services.pipeline_service.pagers import ListPipelineJobsPager


class PipelineJobHook(GoogleBaseHook, OperationHelper):
    """Hook for Google Cloud Vertex AI Pipeline Job APIs."""

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )
        self._pipeline_job: PipelineJob | None = None

    def get_pipeline_service_client(
        self,
        region: str | None = None,
    ) -> PipelineServiceClient:
        """Return PipelineServiceClient object."""
        if region and region != "global":
            client_options = ClientOptions(api_endpoint=f"{region}-aiplatform.googleapis.com:443")
        else:
            client_options = ClientOptions()
        return PipelineServiceClient(
            credentials=self.get_credentials(), client_info=CLIENT_INFO, client_options=client_options
        )

    def get_pipeline_job_object(
        self,
        display_name: str,
        template_path: str,
        job_id: str | None = None,
        pipeline_root: str | None = None,
        parameter_values: dict[str, Any] | None = None,
        input_artifacts: dict[str, str] | None = None,
        enable_caching: bool | None = None,
        encryption_spec_key_name: str | None = None,
        labels: dict[str, str] | None = None,
        project: str | None = None,
        location: str | None = None,
        failure_policy: str | None = None,
    ) -> PipelineJob:
        """Return PipelineJob object."""
        return PipelineJob(
            display_name=display_name,
            template_path=template_path,
            job_id=job_id,
            pipeline_root=pipeline_root,
            parameter_values=parameter_values,
            input_artifacts=input_artifacts,
            enable_caching=enable_caching,
            encryption_spec_key_name=encryption_spec_key_name,
            labels=labels,
            credentials=self.get_credentials(),
            project=project,
            location=location,
            failure_policy=failure_policy,
        )

    def cancel_pipeline_job(self) -> None:
        """Cancel PipelineJob."""
        if self._pipeline_job:
            self._pipeline_job.cancel()

    @GoogleBaseHook.fallback_to_default_project_id
    def create_pipeline_job(
        self,
        project_id: str,
        region: str,
        pipeline_job: PipelineJob,
        pipeline_job_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.PipelineJob:
        """
        Create a PipelineJob. A PipelineJob will run immediately when created.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param pipeline_job:  Required. The PipelineJob to create.
        :param pipeline_job_id:  The ID to use for the PipelineJob, which will become the final component of
            the PipelineJob name. If not provided, an ID will be automatically generated.

            This value should be less than 128 characters, and valid characters are /[a-z][0-9]-/.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_pipeline_service_client(region)
        parent = client.common_location_path(project_id, region)

        result = client.create_pipeline_job(
            request={
                "parent": parent,
                "pipeline_job": pipeline_job,
                "pipeline_job_id": pipeline_job_id,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def run_pipeline_job(
        self,
        project_id: str,
        region: str,
        display_name: str,
        template_path: str,
        job_id: str | None = None,
        pipeline_root: str | None = None,
        parameter_values: dict[str, Any] | None = None,
        input_artifacts: dict[str, str] | None = None,
        enable_caching: bool | None = None,
        encryption_spec_key_name: str | None = None,
        labels: dict[str, str] | None = None,
        failure_policy: str | None = None,
        # START: run param
        service_account: str | None = None,
        network: str | None = None,
        create_request_timeout: float | None = None,
        experiment: str | experiment_resources.Experiment | None = None,
        # END: run param
    ) -> PipelineJob:
        """
        Create and run a PipelineJob until its completion.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param display_name: Required. The user-defined name of this Pipeline.
        :param template_path: Required. The path of PipelineJob or PipelineSpec JSON or YAML file. It can be
            a local path, a Google Cloud Storage URI (e.g. "gs://project.name"), an Artifact Registry URI
            (e.g. "https://us-central1-kfp.pkg.dev/proj/repo/pack/latest"), or an HTTPS URI.
        :param job_id: Optional. The unique ID of the job run. If not specified, pipeline name + timestamp
            will be used.
        :param pipeline_root: Optional. The root of the pipeline outputs. If not set, the staging bucket set
            in aiplatform.init will be used. If that's not set a pipeline-specific artifacts bucket will be
            used.
        :param parameter_values: Optional. The mapping from runtime parameter names to its values that
            control the pipeline run.
        :param input_artifacts: Optional. The mapping from the runtime parameter name for this artifact to
            its resource id. For example: "vertex_model":"456". Note: full resource name
            ("projects/123/locations/us-central1/metadataStores/default/artifacts/456") cannot be used.
        :param enable_caching: Optional. Whether to turn on caching for the run.
            If this is not set, defaults to the compile time settings, which are True for all tasks by
            default, while users may specify different caching options for individual tasks.
            If this is set, the setting applies to all tasks in the pipeline. Overrides the compile time
            settings.
        :param encryption_spec_key_name: Optional. The Cloud KMS resource identifier of the customer managed
            encryption key used to protect the job. Has the form:
            ``projects/my-project/locations/my-region/keyRings/my-kr/cryptoKeys/my-key``.
            The key needs to be in the same region as where the compute resource is created. If this is set,
            then all resources created by the PipelineJob will be encrypted with the provided encryption key.
            Overrides encryption_spec_key_name set in aiplatform.init.
        :param labels: Optional. The user defined metadata to organize PipelineJob.
        :param failure_policy: Optional. The failure policy - "slow" or "fast". Currently, the default of a
            pipeline is that the pipeline will continue to run until no more tasks can be executed, also
            known as PIPELINE_FAILURE_POLICY_FAIL_SLOW (corresponds to "slow"). However, if a pipeline is set
            to PIPELINE_FAILURE_POLICY_FAIL_FAST (corresponds to "fast"), it will stop scheduling any new
            tasks when a task has failed. Any scheduled tasks will continue to completion.
        :param service_account: Optional. Specifies the service account for workload run-as account. Users
            submitting jobs must have act-as permission on this run-as account.
        :param network: Optional. The full name of the Compute Engine network to which the job should be
            peered. For example, projects/12345/global/networks/myVPC.
            Private services access must already be configured for the network. If left unspecified, the
            network set in aiplatform.init will be used. Otherwise, the job is not peered with any network.
        :param create_request_timeout: Optional. The timeout for the create request in seconds.
        :param experiment: Optional. The Vertex AI experiment name or instance to associate to this
            PipelineJob. Metrics produced by the PipelineJob as system.Metric Artifacts will be associated as
            metrics to the current Experiment Run. Pipeline parameters will be associated as parameters to
            the current Experiment Run.
        """
        self._pipeline_job = self.get_pipeline_job_object(
            display_name=display_name,
            template_path=template_path,
            job_id=job_id,
            pipeline_root=pipeline_root,
            parameter_values=parameter_values,
            input_artifacts=input_artifacts,
            enable_caching=enable_caching,
            encryption_spec_key_name=encryption_spec_key_name,
            labels=labels,
            project=project_id,
            location=region,
            failure_policy=failure_policy,
        )
        self._pipeline_job.submit(
            service_account=service_account,
            network=network,
            create_request_timeout=create_request_timeout,
            experiment=experiment,
        )
        self._pipeline_job.wait()

        return self._pipeline_job

    @GoogleBaseHook.fallback_to_default_project_id
    def submit_pipeline_job(
        self,
        project_id: str,
        region: str,
        display_name: str,
        template_path: str,
        job_id: str | None = None,
        pipeline_root: str | None = None,
        parameter_values: dict[str, Any] | None = None,
        input_artifacts: dict[str, str] | None = None,
        enable_caching: bool | None = None,
        encryption_spec_key_name: str | None = None,
        labels: dict[str, str] | None = None,
        failure_policy: str | None = None,
        # START: run param
        service_account: str | None = None,
        network: str | None = None,
        create_request_timeout: float | None = None,
        experiment: str | experiment_resources.Experiment | None = None,
        # END: run param
    ) -> PipelineJob:
        """
        Create and start a PipelineJob run.

        For more info about the client method please see:
        https://cloud.google.com/python/docs/reference/aiplatform/latest/google.cloud.aiplatform.PipelineJob#google_cloud_aiplatform_PipelineJob_submit

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param display_name: Required. The user-defined name of this Pipeline.
        :param template_path: Required. The path of PipelineJob or PipelineSpec JSON or YAML file. It can be
            a local path, a Google Cloud Storage URI (e.g. "gs://project.name"), an Artifact Registry URI
            (e.g. "https://us-central1-kfp.pkg.dev/proj/repo/pack/latest"), or an HTTPS URI.
        :param job_id: Optional. The unique ID of the job run. If not specified, pipeline name + timestamp
            will be used.
        :param pipeline_root: Optional. The root of the pipeline outputs. If not set, the staging bucket set
            in aiplatform.init will be used. If that's not set a pipeline-specific artifacts bucket will be
            used.
        :param parameter_values: Optional. The mapping from runtime parameter names to its values that
            control the pipeline run.
        :param input_artifacts: Optional. The mapping from the runtime parameter name for this artifact to
            its resource id. For example: "vertex_model":"456". Note: full resource name
            ("projects/123/locations/us-central1/metadataStores/default/artifacts/456") cannot be used.
        :param enable_caching: Optional. Whether to turn on caching for the run.
            If this is not set, defaults to the compile time settings, which are True for all tasks by
            default, while users may specify different caching options for individual tasks.
            If this is set, the setting applies to all tasks in the pipeline. Overrides the compile time
            settings.
        :param encryption_spec_key_name: Optional. The Cloud KMS resource identifier of the customer managed
            encryption key used to protect the job. Has the form:
            ``projects/my-project/locations/my-region/keyRings/my-kr/cryptoKeys/my-key``.
            The key needs to be in the same region as where the compute resource is created. If this is set,
            then all resources created by the PipelineJob will be encrypted with the provided encryption key.
            Overrides encryption_spec_key_name set in aiplatform.init.
        :param labels: Optional. The user defined metadata to organize PipelineJob.
        :param failure_policy: Optional. The failure policy - "slow" or "fast". Currently, the default of a
            pipeline is that the pipeline will continue to run until no more tasks can be executed, also
            known as PIPELINE_FAILURE_POLICY_FAIL_SLOW (corresponds to "slow"). However, if a pipeline is set
            to PIPELINE_FAILURE_POLICY_FAIL_FAST (corresponds to "fast"), it will stop scheduling any new
            tasks when a task has failed. Any scheduled tasks will continue to completion.
        :param service_account: Optional. Specifies the service account for workload run-as account. Users
            submitting jobs must have act-as permission on this run-as account.
        :param network: Optional. The full name of the Compute Engine network to which the job should be
            peered. For example, projects/12345/global/networks/myVPC.
            Private services access must already be configured for the network. If left unspecified, the
            network set in aiplatform.init will be used. Otherwise, the job is not peered with any network.
        :param create_request_timeout: Optional. The timeout for the create request in seconds.
        :param experiment: Optional. The Vertex AI experiment name or instance to associate to this PipelineJob.
            Metrics produced by the PipelineJob as system.Metric Artifacts will be associated as metrics
            to the current Experiment Run. Pipeline parameters will be associated as parameters to
            the current Experiment Run.
        """
        self._pipeline_job = self.get_pipeline_job_object(
            display_name=display_name,
            template_path=template_path,
            job_id=job_id,
            pipeline_root=pipeline_root,
            parameter_values=parameter_values,
            input_artifacts=input_artifacts,
            enable_caching=enable_caching,
            encryption_spec_key_name=encryption_spec_key_name,
            labels=labels,
            project=project_id,
            location=region,
            failure_policy=failure_policy,
        )
        self._pipeline_job.submit(
            service_account=service_account,
            network=network,
            create_request_timeout=create_request_timeout,
            experiment=experiment,
        )

        return self._pipeline_job

    @GoogleBaseHook.fallback_to_default_project_id
    def get_pipeline_job(
        self,
        project_id: str,
        region: str,
        pipeline_job_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.PipelineJob:
        """
        Get a PipelineJob.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param pipeline_job_id: Required. The ID of the PipelineJob resource.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_pipeline_service_client(region)
        name = client.pipeline_job_path(project_id, region, pipeline_job_id)

        result = client.get_pipeline_job(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def list_pipeline_jobs(
        self,
        project_id: str,
        region: str,
        page_size: int | None = None,
        page_token: str | None = None,
        filter: str | None = None,
        order_by: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListPipelineJobsPager:
        """
        List PipelineJobs in a Location.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param filter: Optional. Lists the PipelineJobs that match the filter expression. The
            following fields are supported:

            -  ``pipeline_name``: Supports ``=`` and ``!=`` comparisons.
            -  ``display_name``: Supports ``=``, ``!=`` comparisons, and
               ``:`` wildcard.
            -  ``pipeline_job_user_id``: Supports ``=``, ``!=``
               comparisons, and ``:`` wildcard. for example, can check
               if pipeline's display_name contains *step* by doing
               display_name:"*step*"
            -  ``create_time``: Supports ``=``, ``!=``, ``<``, ``>``,
               ``<=``, and ``>=`` comparisons. Values must be in RFC
               3339 format.
            -  ``update_time``: Supports ``=``, ``!=``, ``<``, ``>``,
               ``<=``, and ``>=`` comparisons. Values must be in RFC
               3339 format.
            -  ``end_time``: Supports ``=``, ``!=``, ``<``, ``>``,
               ``<=``, and ``>=`` comparisons. Values must be in RFC
               3339 format.
            -  ``labels``: Supports key-value equality and key presence.

            Filter expressions can be combined together using logical
            operators (``AND`` & ``OR``). For example:
            ``pipeline_name="test" AND create_time>"2020-05-18T13:30:00Z"``.

            The syntax to define filter expression is based on
            https://google.aip.dev/160.
        :param page_size: Optional. The standard list page size.
        :param page_token: Optional. The standard list page token. Typically obtained via
            [ListPipelineJobsResponse.next_page_token][google.cloud.aiplatform.v1.ListPipelineJobsResponse.next_page_token]
            of the previous
            [PipelineService.ListPipelineJobs][google.cloud.aiplatform.v1.PipelineService.ListPipelineJobs]
            call.
        :param order_by: Optional. A comma-separated list of fields to order by. The default
            sort order is in ascending order. Use "desc" after a field
            name for descending. You can have multiple order_by fields
            provided e.g. "create_time desc, end_time", "end_time,
            start_time, update_time" For example, using "create_time
            desc, end_time" will order results by create time in
            descending order, and if there are multiple jobs having the
            same create time, order them by the end time in ascending
            order. if order_by is not specified, it will order by
            default order is create time in descending order. Supported
            fields:

            -  ``create_time``
            -  ``update_time``
            -  ``end_time``
            -  ``start_time``
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_pipeline_service_client(region)
        parent = client.common_location_path(project_id, region)

        result = client.list_pipeline_jobs(
            request={
                "parent": parent,
                "page_size": page_size,
                "page_token": page_token,
                "filter": filter,
                "order_by": order_by,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_pipeline_job(
        self,
        project_id: str,
        region: str,
        pipeline_job_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Delete a PipelineJob.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param pipeline_job_id: Required. The ID of the PipelineJob resource to be deleted.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_pipeline_service_client(region)
        name = client.pipeline_job_path(project_id, region, pipeline_job_id)

        result = client.delete_pipeline_job(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @staticmethod
    def extract_pipeline_job_id(obj: dict) -> str:
        """Return unique id of a pipeline job from its name."""
        return obj["name"].rpartition("/")[-1]


class PipelineJobAsyncHook(GoogleBaseAsyncHook):
    """Asynchronous hook for Google Cloud Vertex AI Pipeline Job APIs."""

    sync_hook_class = PipelineJobHook
    PIPELINE_COMPLETE_STATES = (
        PipelineState.PIPELINE_STATE_CANCELLED,
        PipelineState.PIPELINE_STATE_FAILED,
        PipelineState.PIPELINE_STATE_PAUSED,
        PipelineState.PIPELINE_STATE_SUCCEEDED,
    )

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )

    async def get_credentials(self) -> Credentials:
        sync_hook = await self.get_sync_hook()
        return sync_hook.get_credentials()

    async def get_project_id(self) -> str:
        sync_hook = await self.get_sync_hook()
        return sync_hook.project_id

    async def get_location(self) -> str:
        sync_hook = await self.get_sync_hook()
        return sync_hook.location

    async def get_pipeline_service_client(
        self,
        region: str | None = None,
    ) -> PipelineServiceAsyncClient:
        if region and region != "global":
            client_options = ClientOptions(api_endpoint=f"{region}-aiplatform.googleapis.com:443")
        else:
            client_options = ClientOptions()
        return PipelineServiceAsyncClient(
            credentials=await self.get_credentials(),
            client_info=CLIENT_INFO,
            client_options=client_options,
        )

    async def get_pipeline_job(
        self,
        project_id: str,
        location: str,
        job_id: str,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | _MethodDefault | None = DEFAULT,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.PipelineJob:
        """
        Get a PipelineJob proto message from PipelineServiceAsyncClient.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param job_id: Required. The ID of the PipelineJob resource.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = await self.get_pipeline_service_client(region=location)
        pipeline_job_name = client.pipeline_job_path(
            project=project_id,
            location=location,
            pipeline_job=job_id,
        )
        pipeline_job: types.PipelineJob = await client.get_pipeline_job(
            request={"name": pipeline_job_name},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return pipeline_job

    async def wait_for_pipeline_job(
        self,
        project_id: str,
        location: str,
        job_id: str,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        poll_interval: int = 10,
    ) -> types.PipelineJob:
        """Wait until the pipeline job is in a complete state and return it."""
        while True:
            try:
                self.log.info("Requesting a pipeline job with id %s", job_id)
                job: types.PipelineJob = await self.get_pipeline_job(
                    project_id=project_id,
                    location=location,
                    job_id=job_id,
                    retry=retry,
                    timeout=timeout,
                    metadata=metadata,
                )
            except Exception as ex:
                self.log.exception("Exception occurred while requesting pipeline job %s", job_id)
                raise AirflowException(ex)

            self.log.info("Status of the pipeline job %s is %s", job.name, job.state.name)
            if job.state in self.PIPELINE_COMPLETE_STATES:
                return job

            self.log.info("Sleeping for %s seconds.", poll_interval)
            await asyncio.sleep(poll_interval)
